// ------------------------------------------------------------------
// ---                                                            ---
// --- PWM to PPM encoder using ATtiny2313                        ---
// ---                                                            ---
// ---                                21.aug.2010, Matej Kogovsek ---
// ---                                                            ---
// --- FUSES: 0xDFE4 (same as factory default but without CKDIV8) ---
// ---                                                            ---
// ------------------------------------------------------------------

// F_CPU 8000000, defined in makefile

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>

#define PIN(x) (*(&x - 2))  // address of input register of port x
#define DDR(x) (*(&x - 1))  // address of data direction register of port x

// ------------------------------------------------------------------
// --- PARAMETERS ---------------------------------------------------
// ------------------------------------------------------------------

#define PPM_PORT PORTD
#define PPM_PIN PD4

#define MAX_WIDTH_USEC 2150 // if a channel pulse is longer than this
#define MIN_WIDTH_USEC 850 // or shorter than this, these values are used
#define END_WIDTH_USEC 2150 // PPM end low pulse width
#define MAX_SAMPLES 20		// max number of pin change samples to remember per run
#define NUM_CH 8			// number of rx channels
#define PPM_PW_USEC 250		// PPM pulse width
#define EOF_TIMEOUT 10 		// assume end of frame when no change for X*256usec (10 => 2560)

// ------------------------------------------------------------------
// --- VARIABLES ----------------------------------------------------
// ------------------------------------------------------------------

volatile uint8_t  pin_stat[MAX_SAMPLES+1];
volatile uint16_t pin_time[MAX_SAMPLES+1];
register uint8_t  pin_samp asm("r2");	// samples counter

volatile uint16_t t1times[NUM_CH+1];

register uint8_t time_hi asm("r3");		// high byte for usec counter, low byte is TCNT0

// ------------------------------------------------------------------
// --- MAIN ---------------------------------------------------------
// ------------------------------------------------------------------

int main(void)
{
	ACSR |= _BV(ACD);					// turn off analog comparator
	
	DDR(PORTB) = 0;						// configure PORTB as input
	DDR(PPM_PORT) |= _BV(PPM_PIN);		// configure ppm pin as output
	
	TIMSK = _BV(TOIE0) | _BV(TOIE1);	// enable overflow interrupts for both timers
	TIFR = _BV(TOV0) | _BV(TOV0);		// clear both overflow interrupt flags
	
	PCMSK = (1<<NUM_CH)-1;				// define pin change interrupt pin mask
	GIMSK = _BV(PCIE);					// enable pin change interrupt
	EIFR = _BV(PCIF);					// clear pin change interrupt flag
	
	pin_samp = 0;
	time_hi = 0;
	
	wdt_enable(WDTO_250MS);				// enable watchdog

	sei();								// enable interrupts

	while( 1 ) {
		asm("nop");						// no code here so interrupts can be naked
	}
}

// ------------------------------------------------------------------
// --- INTERRUPTS ---------------------------------------------------
// ------------------------------------------------------------------

// pin change interrupt - records PINB status and time to memory
ISR(SIG_PIN_CHANGE, ISR_NAKED)
{
	asm("clr __zero_reg__");
	uint8_t b = PINB;			// store pin status
	
	TCCR0B = 0;					// stop Timer0
	uint16_t t;
	if( TIFR & _BV(TOV0) )	{	// if Timer0 overflowed (and ovf int wasn't yet executed)
		t = (time_hi+1) << 8;
	} else {
		t = (time_hi << 8) + TCNT0;
	}
	TCNT0 = 0;					// zero Timer0 counter
	time_hi = 0;				// and time_hi
	TIFR = _BV(TOV0);			// clear possible pending Timer0 overflow interrupt
	TCCR0B = _BV(CS01);			// start Timer0
	
	pin_time[pin_samp] = t;		// store time (pin_time[0] is ignored)
	pin_stat[pin_samp] = b;		// and pin status
	
	if( pin_samp < MAX_SAMPLES ) { // "array index out of bounds" guard
		pin_samp++;
	}
	asm("reti");
}

// Timer0 overflow - keeps time and checks for no pin change timeout (frame end)
// timer should be prescaled to increase every 1us
ISR(SIG_OVERFLOW0, ISR_NAKED)
{
	asm("clr __zero_reg__");
	time_hi++;
	
	if( time_hi == EOF_TIMEOUT ) {	// end of frame, setup Timer1 and start it
		TCCR0B = 0;					// disable Timer0
		
		if ( TCCR1B == 0 ) { 		// if Timer1 is not already running
			uint8_t ch;
			
			for( ch = 0; ch < NUM_CH; ch++) {	// setup Timer1 times, loop through channels
				uint16_t t = 0;
				uint8_t i;
				uint8_t b = 0;
				
				for( i = 1; i < pin_samp; i++ ) { // loop through samples
					if( pin_stat[i-1] & _BV(ch) ) {
						t += pin_time[i];
						b = 1;					// channel up was detected
					                              } 
					else {
						if( b ) { break; } 		// channel down, but was up, break (speedup)
					        }
				    }
				
				// sanity check
				if( t > MAX_WIDTH_USEC ) { t = MAX_WIDTH_USEC; } else
				if( t < MIN_WIDTH_USEC ) { t = MIN_WIDTH_USEC; }
				t1times[ch] = t;
			}
			t1times[NUM_CH] = END_WIDTH_USEC;	// end pulse
			
			TCNT1 = 0xffff;			// set TCNT1 so it overflows immediately
			TCCR1B = _BV(CS11);		// enable Timer1 (start PPM pulse train transmission)
		}
		
		pin_samp = 0;				// clear pin change memory
	}
	asm("reti");
}

// Timer1 overflow - sets PPM pin
// timer should be prescaled to increase every 1us
// overflow time is set with starting TCNT1 value
ISR(SIG_OVERFLOW1, ISR_NAKED)
{
	asm("clr __zero_reg__");
	static uint8_t t1state = 0;

	if( t1state & 1 ) {
		// odd states - ppm goes low for duration of PWM pulse minus PPM_PW_USEC
		PPM_PORT &= ~_BV(PPM_PIN);  		// ppm low
		TCNT1 = 0xffff - t1times[t1state >> 1] + PPM_PW_USEC;
	} else {
		// even states - ppm goes high for PPM_PW_USEC
		PPM_PORT |= _BV(PPM_PIN);   		// ppm high
		TCNT1 = 0xffff - PPM_PW_USEC;
	}

	if( t1state == (NUM_CH+NUM_CH+1) ) {	// end of sequence
		t1state = 0;
		TCCR1B = 0;       					// disable Timer1
	wdt_reset();
	} else {
		t1state++;
	}
	asm("reti");
}

// ------------------------------------------------------------------
